#ifndef GENERICKDTREE_H
#define GENERICKDTREE_H

#include <vector>
#include <algorithm>
#include <cmath>
#include <limits>
#include "vec.h"

template<typename point_type>
struct Trait
{
	typedef float	scalar_type;

	enum { dimensions = 3 };

	static inline const scalar_type &get(const point_type &p, const size_t idx)
	{
		return p[idx];
	}

	static inline scalar_type &get(point_type &p, const size_t idx)
	{
		return p[idx];
	}
};

template<typename T, int N>
struct Trait<T[N]>
{
	typedef T	scalar_type;

	enum { dimensions = N };

	static inline const scalar_type &get(const T p[N], const size_t idx)
	{
		return p[idx];
	}

	static inline scalar_type &get(T p[N], const size_t idx)
	{
		return p[idx];
	}
};

template<typename T, int N>
struct Trait<Vec<T, N> >
{
	typedef T	scalar_type;

	enum { dimensions = N };

	static inline const scalar_type &get(const Vec<T, N> &p, const size_t idx)
	{
		return p[idx];
	}

	static inline scalar_type &get(Vec<T, N> &p, const size_t idx)
	{
		return p[idx];
	}
};

template<typename K, typename V>
struct Trait<std::pair<K, V> > : public Trait<K>
{
	static inline const typename Trait<K>::scalar_type &get(const std::pair<K, V> &p, const size_t idx)
	{
		return Trait<K>::get(p.first, idx);
	}

	static inline typename Trait<K>::scalar_type &get(std::pair<K, V> &p, const size_t idx)
	{
		return Trait<K>::get(p.first, idx);
	}
};

struct N1
{
	template<typename point_type>
	static inline typename Trait<point_type>::scalar_type norm(const point_type &p)
	{
		typename Trait<point_type>::scalar_type acc(0);
		for(size_t i = 0 ; i < Trait<point_type>::dimensions ; ++i)
			acc += std::fabs(Trait<point_type>::get(p, i));
		return acc;
	}

	template<typename point_type1, typename point_type2>
	static inline typename Trait<point_type1>::scalar_type distance(const point_type1 &p, const point_type2 &q)
	{
		typename Trait<point_type1>::scalar_type acc(0);
		for(size_t i = 0 ; i < Trait<point_type1>::dimensions ; ++i)
			acc += std::fabs(Trait<point_type1>::get(p, i) - Trait<point_type2>::get(q, i));
		return acc;
	}
};

struct N2
{
	template<typename point_type>
	static inline typename Trait<point_type>::scalar_type norm(const point_type &p)
	{
		typename Trait<point_type>::scalar_type acc(0);
		for(size_t i = 0 ; i < Trait<point_type>::dimensions ; ++i)
		{
			typename Trait<point_type>::scalar_type v = Trait<point_type>::get(p, i);
			acc += v * v;
		}
		return std::sqrt(acc);
	}

	template<typename point_type1, typename point_type2>
	static inline typename Trait<point_type1>::scalar_type distance(const point_type1 &p, const point_type2 &q)
	{
		typename Trait<point_type1>::scalar_type acc(0);
		for(size_t i = 0 ; i < Trait<point_type1>::dimensions ; ++i)
		{
			typename Trait<point_type1>::scalar_type v = Trait<point_type1>::get(p, i) - Trait<point_type2>::get(q, i);
			acc += v * v;
		}
		return std::sqrt(acc);
	}
};

template<typename point_type, typename Norm = N1>
class GenericKdTree
{
	typedef typename Trait<point_type>::scalar_type	scalar_type;
	enum { D = Trait<point_type>::dimensions };
public:
	GenericKdTree(typename std::vector<point_type> &elts)
		: childs(NULL)
	{
		build(elts.begin(), elts.end());
	}

	~GenericKdTree()
	{
		if (childs)
			delete[] childs;
	}

	template<typename PT, typename Fn>
	void foreach_neighbors(const PT &q, const scalar_type r, Fn f) const
	{
		if (Norm::distance(q, center) > r + this->r)
			return;

		if (!childs)
		{
			for(typename std::vector<point_type>::iterator it = begin ; it < end ; ++it)
				if (Norm::distance(q, *it) <= r)
					f(*it);
			return;
		}

		const scalar_type s = Trait<PT>::get(q, sep) - center[sep];

		if (s < scalar_type(0))
		{
			childs[0].foreach_neighbors(q, r, f);
			if (-s <= r)
				childs[1].foreach_neighbors(q, r, f);
		}
		else
		{
			childs[1].foreach_neighbors(q, r, f);
			if (s <= r)
				childs[0].foreach_neighbors(q, r, f);
		}
	}

	template<typename PT>
	point_type *nearest(const PT &q, scalar_type *r = NULL) const
	{
		if (r && Norm::distance(q, center) > *r + this->r)
			return NULL;

		if (!childs)
		{
			scalar_type min_dist = r ? *r : std::numeric_limits<scalar_type>::max();
			point_type *best = NULL;
			for(typename std::vector<point_type>::iterator it = begin ; it < end ; ++it)
			{
				const scalar_type distance = Norm::distance(q, *it);
				if (distance < min_dist)
				{
					min_dist = distance;
					best = &(*it);
				}
			}
			if (r && best)
				*r = min_dist;
			return best;
		}

		const scalar_type s = Trait<PT>::get(q, sep) - center[sep];

		if (s < scalar_type(0))
		{
			scalar_type min_dist = r ? *r : std::numeric_limits<scalar_type>::max();
			point_type *best = childs[0].nearest(q, &min_dist);
			if (!best)
				return childs[1].nearest(q, r);
			if (-s <= min_dist)
			{
				point_type *candidate = childs[1].nearest(q, &min_dist);
				if (candidate)
					best = candidate;
			}
			if (r && best)
				*r = min_dist;
			return best;
		}
		else
		{
			scalar_type min_dist = r ? *r : std::numeric_limits<scalar_type>::max();
			point_type *best = childs[1].nearest(q, &min_dist);
			if (!best)
				return childs[0].nearest(q, r);
			if (s <= min_dist)
			{
				point_type *candidate = childs[0].nearest(q, &min_dist);
				if (candidate)
					best = candidate;
			}
			if (r && best)
				*r = min_dist;
			return best;
		}
	}

	template<typename PT, typename Fn>
	point_type *nearest_if(const PT &q, Fn f, scalar_type *r = NULL) const
	{
		if (r && Norm::distance(q, center) > *r + this->r)
			return NULL;

		if (!childs)
		{
			scalar_type min_dist = r ? *r : std::numeric_limits<scalar_type>::max();
			point_type *best = NULL;
			for(typename std::vector<point_type>::iterator it = begin ; it < end ; ++it)
			{
				const scalar_type distance = Norm::distance(q, *it);
				if (distance < min_dist && f(*it))
				{
					min_dist = distance;
					best = &(*it);
				}
			}
			if (r && best)
				*r = min_dist;
			return best;
		}

		const scalar_type s = Trait<PT>::get(q, sep) - center[sep];

		if (s < scalar_type(0))
		{
			scalar_type min_dist = r ? *r : std::numeric_limits<scalar_type>::max();
			point_type *best = childs[0].nearest_if(q, f, &min_dist);
			if (!best)
				return childs[1].nearest_if(q, f, r);
			if (-s < min_dist)
			{
				point_type *candidate = childs[1].nearest_if(q, f, &min_dist);
				if (candidate)
					best = candidate;
			}
			if (r && best)
				*r = min_dist;
			return best;
		}
		else
		{
			scalar_type min_dist = r ? *r : std::numeric_limits<scalar_type>::max();
			point_type *best = childs[1].nearest_if(q, f, &min_dist);
			if (!best)
				return childs[0].nearest_if(q, f, r);
			if (s < min_dist)
			{
				point_type *candidate = childs[0].nearest_if(q, f, &min_dist);
				if (candidate)
					best = candidate;
			}
			if (r && best)
				*r = min_dist;
			return best;
		}
	}

private:
	GenericKdTree() : childs(NULL)	{}

	void build(typename std::vector<point_type>::iterator begin,
			   typename std::vector<point_type>::iterator end)
	{
		childs = NULL;
		this->begin = begin;
		this->end = end;
		const size_t size = end - begin;
		if (size == 0)
			return;

		scalar_type	lower[D], upper[D];
		for(size_t i = 0 ; i < D ; ++i)
			lower[i] = upper[i] = Trait<point_type>::get(*begin, i);
		for(typename std::vector<point_type>::iterator it = begin ; it < end ; ++it)
			for(size_t i = 0 ; i < D ; ++i)
			{
				const scalar_type v = Trait<point_type>::get(*it, i);
				lower[i] = std::min(v, lower[i]);
				upper[i] = std::max(v, upper[i]);
			}
		for(size_t i = 0 ; i < D ; ++i)
			center[i] = (upper[i] + lower[i]) / 2;
		r = Norm::distance(upper, lower) / 2;
		if (size < 8)
			return;

		sep = 0;
		for(size_t i = 1 ; i < D ; ++i)
			if (upper[i] - lower[i] > upper[sep] - lower[sep])
				sep = i;

		typename std::vector<point_type>::iterator mid = std::partition(begin, end, [&](const point_type &p) { return Trait<point_type>::get(p, sep) < center[sep]; });
		if (mid == begin || mid == end)
			return;

		childs = new GenericKdTree[2];
		childs[0].build(begin, mid);
		childs[1].build(mid, end);
	}

private:
	typename std::vector<point_type>::iterator	begin, end;
	GenericKdTree *childs;
	scalar_type	center[D];
	int sep;
	scalar_type r;
};

template<typename point_type, typename Norm = N1>
inline void space_partition(typename std::vector<point_type>::iterator begin,
							typename std::vector<point_type>::iterator end)
{
	const size_t size = end - begin;
	if (size <= 8)
		return;

	typedef typename Trait<point_type>::scalar_type scalar_type;
	enum { D = Trait<point_type>::dimensions };

	scalar_type	lower[D], upper[D];
	for(size_t i = 0 ; i < D ; ++i)
		lower[i] = upper[i] = Trait<point_type>::get(*begin, i);
	for(typename std::vector<point_type>::const_iterator it = begin ; it < end ; ++it)
		for(size_t i = 0 ; i < D ; ++i)
		{
			const scalar_type v = Trait<point_type>::get(*it, i);
			lower[i] = std::min(v, lower[i]);
			upper[i] = std::max(v, upper[i]);
		}
	int sep = 0;
	for(size_t i = 1 ; i < D ; ++i)
		if (upper[i] - lower[i] > upper[sep] - lower[sep])
			sep = i;
	const scalar_type center = (upper[sep] + lower[sep]) / 2;

	typename std::vector<point_type>::iterator mid = std::partition(begin, end, [sep,center](const point_type &p) { return Trait<point_type>::get(p, sep) < center; });
	if (mid == begin || mid == end)
		return;

	space_partition<point_type, Norm>(begin, mid);
	space_partition<point_type, Norm>(mid, end);
}

#endif
