#include "serialconsole.h"
#include "simpletextedit.h"
#include <QLineEdit>
#include <QVBoxLayout>
#include <QDebug>
#include <QTimer>
#include <QTextStream>
#include <QComboBox>
#include <QtSerialPort/QSerialPortInfo>
#include <QSettings>
#include <QScrollBar>
#include <QHBoxLayout>
#include "delay.h"

SerialConsole *SerialConsole::pInstance = NULL;

SerialConsole *SerialConsole::instance()
{
    if (!pInstance)
        pInstance = new SerialConsole;
    return pInstance;
}

SerialConsole::SerialConsole(QWidget *parent) :
    QDockWidget(tr("Serial console"), parent)
{
	setObjectName("SerialConsole");

    extruder_temperature = 0.0;
    bed_temperature = 0.0;
    nb_processed_commands = 0;
    start_time.start();

    QWidget *widget = new QWidget;
    setWidget(widget);
    QVBoxLayout *layout = new QVBoxLayout;
    widget->setLayout(layout);

    QComboBox *serial_ports = new QComboBox;

	const QList<QSerialPortInfo> &available_ports = QSerialPortInfo::availablePorts();
	for(const QSerialPortInfo &port : available_ports)
    {
        serial_ports->addItem(QIcon::fromTheme("printer"), port.portName(), port.portName());
    }
    connect(serial_ports,SIGNAL(activated(QString)),this,SLOT(connectTo(QString)));

    QHBoxLayout *history_layout = new QHBoxLayout;
    scroll_bar = new QScrollBar(Qt::Vertical);
    scroll_bar->setMinimum(0);
    scroll_bar->setValue(0);
    history_layout->addWidget(history = new SimpleTextEdit);
	history->setMaxSize(10000);
    history_layout->addWidget(scroll_bar);
    connect(history,SIGNAL(scrollPositionChanged(int)),scroll_bar,SLOT(setValue(int)));
    connect(scroll_bar,SIGNAL(valueChanged(int)),history,SLOT(setScrollPosition(int)));
    connect(history,SIGNAL(maxScrollChanged(int)),this,SLOT(updateScrollBar(int)));

    layout->addWidget(serial_ports);
    layout->addLayout(history_layout);
    layout->addWidget(cmd = new QLineEdit);

    history->setFocusPolicy(Qt::NoFocus);
    cmd->setFocusPolicy(Qt::StrongFocus);

	rs232 = new QSerialPort(this);
	rs232->setBaudRate(115200);

    connect(rs232,SIGNAL(readyRead()),this,SLOT(readMessage()));
    connect(cmd,SIGNAL(returnPressed()),this,SLOT(sendMessage()));

    ack_timer = new QTimer(this);
    ack_timer->setSingleShot(true);
    ack_timer->setInterval(2000);
    connect(ack_timer,SIGNAL(timeout()),this,SLOT(kickAfterTimeout()));

    QTimer *timer = new QTimer(this);
    timer->setInterval(10000);
    timer->setSingleShot(false);
    connect(timer,SIGNAL(timeout()),this,SLOT(requestTemperature()));
    timer->start();

    ack_expected_before_sending_a_command = 0;
    b_paused = false;

    QSettings settings;
    settings.beginGroup("SerialConsole");
    const QString &serial_port = settings.value("serial_port", QString()).toString();
    settings.endGroup();

    int idx = 0;
	for(const QSerialPortInfo &port : available_ports)
        if (port.portName() == serial_port)
        {
            serial_ports->setCurrentIndex(idx);
            connectTo(serial_port);
            break;
        }
        else
            ++idx;
}

void SerialConsole::readMessage()
{
    if (!rs232->isOpen())
        return;
    const double old_extruder_temperature = extruder_temperature;
    const double old_bed_temperature = bed_temperature;

    static const QRegExp sep("[\\s,]");

    while(rs232->canReadLine())
    {
        const QByteArray &msg = rs232->readLine().trimmed();
        if (msg.isEmpty())
            continue;
        history->append("-> " + msg);

        const QList<QString> &args = QString(msg).split(sep, QString::SkipEmptyParts);

        for(int i = 0 ; i < args.size() ; ++i)
        {
            const QString &arg = args[i];
            if (arg == "ok")
            {
                ack_timer->stop();
                ack_expected_before_sending_a_command--;
                if (!command_queue.isEmpty() && ack_expected_before_sending_a_command == 0)
                {
                    const QByteArray msg = command_queue.front();
                    command_queue.pop_front();
                    if (nb_processed_commands == 0)
                        start_time.start();
                    ++nb_processed_commands;
                    sendMessage(msg);
                }
                else if (command_queue.isEmpty())
                {
                    nb_processed_commands = 0;
                    emit updateWorkProgress(100);
                    emit updateETA(0.0);
                }
            }
            else if (arg.startsWith("X:"))
            {
                emit updateCoordinateX(arg.right(arg.size() - 2).toDouble());
            }
            else if (arg.startsWith("Y:"))
            {
                emit updateCoordinateY(arg.right(arg.size() - 2).toDouble());
            }
            else if (arg.startsWith("Z:"))
            {
                emit updateCoordinateZ(arg.right(arg.size() - 2).toDouble());
            }
            else if (arg == "rs" && i + 1 < args.size())
            {
                const QString &arg1 = args[i + 1];
                if (arg1.startsWith('N'))
                {
                    ++i;
                    const int line_number = arg1.right(arg1.size() - 1).toInt();
                    if (line_number >= 0 && line_number < command_history.size())
                    {
                        history->append("<- " + command_history[line_number].trimmed());
                        rs232->write(command_history[line_number]);
                        if (ack_timer->isActive())
                        {
                            ack_timer->stop();
                            ack_timer->start();
                        }
                    }
                    else
                    {
                        history->append("<- ");
                        rs232->write("\n");
                    }
                    if (ack_timer->isActive())
                    {
                        ack_timer->stop();
                        ack_timer->start();
                    }
                }
            }
            else if (arg.startsWith("T:"))
            {
                extruder_temperature = arg.right(arg.size() - 2).toDouble();
            }
            else if (arg.startsWith("B:"))
            {
                bed_temperature = arg.right(arg.size() - 2).toDouble();
            }
        }
    }
    if (extruder_temperature != old_extruder_temperature
            || bed_temperature != old_bed_temperature)
        emit updateTemperature(extruder_temperature, bed_temperature);
}

void SerialConsole::sendMessage()
{
    const QString msg = cmd->text();
    cmd->setText(QString());
    sendMessage(msg.toLatin1());
}

void SerialConsole::kickAfterTimeout()
{
    if (!rs232->isOpen() || last_cmd.isEmpty())
        return;
    ack_timer->stop();
    ack_expected_before_sending_a_command = 0;
    if (!command_queue.isEmpty())
    {
        const QByteArray msg = command_queue.front();
        command_queue.pop_front();
        if (nb_processed_commands == 0)
            start_time.start();
        ++nb_processed_commands;
        sendMessage(msg);
    }
    else if (command_queue.isEmpty())
    {
        nb_processed_commands = 0;
        emit updateWorkProgress(100);
        emit updateETA(0.0);
    }
}

void SerialConsole::sendMessage(const QByteArray &msg)
{
    if (msg.isEmpty() || msg.isNull() || !rs232->isOpen())
        return;
    if (ack_expected_before_sending_a_command == 0 && !b_paused)
    {
        if (msg.startsWith("M110"))
            command_history.clear();
        QByteArray checkable_message = 'N' + QByteArray::number(command_history.size()) + ' ' + msg;
        int h(0);
        for(int i = 0 ; i < checkable_message.size() ; ++i)
           h = h ^ checkable_message[i];
        h &= 0xFF;
        checkable_message += '*' + QByteArray::number(h) + '\n';
        command_history << checkable_message;

        while(rs232->bytesToWrite() > 0)
            rs232->waitForBytesWritten(10);
        rs232->write(checkable_message);
        while(rs232->bytesToWrite() > 0)
            rs232->waitForBytesWritten(10);
//        rs232->flush();
        Delay::msleep(10);
        last_cmd = checkable_message;
        ack_expected_before_sending_a_command = 1;
        if (!msg.startsWith("G28"))
            ack_timer->start();
        else
            last_cmd.clear();
        history->append("<- " + checkable_message.trimmed());
    }
    else
    {
        command_queue.push_back(msg);
    }
    if (nb_processed_commands + command_queue.size() == 0)
    {
        emit updateWorkProgress(100);
        emit updateETA(0.0);
    }
    else
    {
        emit updateWorkProgress(nb_processed_commands * 100 / (nb_processed_commands + command_queue.size()));
        if (nb_processed_commands > 0)
            emit updateETA(double(start_time.elapsed()) * 0.001 * command_queue.size() / nb_processed_commands);
    }
}

void SerialConsole::requestTemperature()
{
    if (!rs232->isOpen())
        return;
    if (ack_expected_before_sending_a_command == 0)
    {
        sendMessage("M105");
    }
    else
        command_queue.push_front("M105");
}

void SerialConsole::emergencyStop()
{
    b_paused = false;
    ack_timer->stop();
    ack_expected_before_sending_a_command = 0;
    command_queue.clear();
    last_cmd.clear();
    sendMessage("M0");
    nb_processed_commands = 0;
    emit updateWorkProgress(100);
}

void SerialConsole::stop()
{
    b_paused = false;
    ack_timer->stop();
    ack_expected_before_sending_a_command = 0;
    last_cmd.clear();
    command_queue.clear();
    nb_processed_commands = 0;
    emit updateWorkProgress(100);
}

void SerialConsole::pause()
{
    b_paused = true;
    ack_timer->stop();
}

void SerialConsole::unpause()
{
    b_paused = false;
    if (ack_expected_before_sending_a_command > 0)
    {
        ack_timer->start();
    }
    else if (!command_queue.isEmpty())
    {
        const QByteArray cmd = command_queue.front();
        command_queue.pop_front();
        sendMessage(cmd);
    }
}

void SerialConsole::step()
{
    if (!b_paused)
        return;
    b_paused = false;
    if (ack_expected_before_sending_a_command > 0)
    {
        ack_timer->stop();
        ack_timer->start();
    }
    else if (!command_queue.isEmpty())
    {
        const QByteArray cmd = command_queue.front();
        command_queue.pop_front();
        ++nb_processed_commands;
        sendMessage(cmd);
    }
    b_paused = true;
}

void SerialConsole::connectTo(const QString &portname)
{
    ack_timer->stop();
    last_cmd.clear();
    b_paused = false;
    ack_expected_before_sending_a_command = 0;
    command_queue.clear();

    rs232->close();
    rs232->setPortName(portname);
    if (!rs232->open(QIODevice::ReadWrite))
        history->append("error connecting serial port\n");
    else
    {
        rs232->setDataBits(QSerialPort::Data8);
        rs232->setParity(QSerialPort::NoParity);
        rs232->setStopBits(QSerialPort::OneStop);
        rs232->setFlowControl(QSerialPort::HardwareControl);

        disconnect(rs232,SIGNAL(readyRead()),this,SLOT(readMessage()));

        QList<qint32> baudrates = QSerialPortInfo::standardBaudRates();
        if (!baudrates.contains(250000))
            baudrates.append(250000);
        bool b_valid = false;
        for(int i = baudrates.size() - 1 ; i >= 0 && !b_valid ; --i)
        {
            qDebug() << "trying " << baudrates[i] << " bauds";
            rs232->flush();
            rs232->readAll();
            rs232->setBaudRate(baudrates[i]);
            rs232->write("\nN0 M110*35\n");
            rs232->flush();
            rs232->waitForBytesWritten(10);
            QTime timer;
            timer.start();
            do
            {
                rs232->waitForReadyRead(10);
                while(rs232->canReadLine())
                {
                    QByteArray line = rs232->readLine().trimmed();
                    b_valid |= line == "ok";
                }
            } while (rs232->bytesAvailable() > 0 && timer.elapsed() < 100);
            if (b_valid)
                qDebug() << "working baud rate detected : " << baudrates[i];
        }

        if (b_valid)
            connect(rs232,SIGNAL(readyRead()),this,SLOT(readMessage()));

        history->append("connected on " + portname.toLatin1());
        QSettings settings;
        settings.beginGroup("SerialConsole");
        settings.setValue("serial_port", portname);
        settings.endGroup();

        command_history.clear();
        sendMessage("M110");

        nb_processed_commands = 0;
        emit updateWorkProgress(100);
    }
}

void SerialConsole::updateScrollBar(int s)
{
    scroll_bar->setMaximum(s);
}
